/**
 * @file xrew.cc
 *
 * @brief implements xrew routines.
 */

#include <map>

#include <cassert>
#include <cstdlib>
#include <cstring>

#include "xrew.h"
#include "../hfst.h"

char *_xrew_data;
size_t _xrew_len;
HWFST::TransducerHandle _xrew_transducer_;
HWFST::KeySet* _xrew_keys;
HWFST::KeySet* _xrew_negation_sigma;
HWFST::KeySet* _xrew_creation_sigma;
HWFST::KeyTable* _xrew_key_table;
std::map<HWFST::Key,HWFST::TransducerHandle> _xrew_definitions;
bool _xrew_initialised;
bool _xrew_weighted;
extern char* xrewtext;

int
xrewerror(char *msg)
{
#ifndef NDEBUG
	fprintf(stderr, "*** XRE PARSE ERROR: %s\n", msg);
	if (strlen(_xrew_data) < 60)
	{
		fprintf(stderr, "***	parsing %s [near %s]\n", _xrew_data, xrewtext);
	}
	else
	{
		fprintf(stderr, "***	parsing %60s [near %s]...\n", 
				_xrew_data, xrewtext);
	}
#endif
	return 0;
}

int
xrew_getinput(char *buf, int maxlen)
{
	int retval = 0;
	if ( maxlen > _xrew_len ) {
		maxlen = _xrew_len;
	}
	memcpy(buf, _xrew_data, maxlen);
	_xrew_data += maxlen;
	_xrew_len -= maxlen;
	retval = maxlen;
	return retval;
}

char*
xrew_strip_percents(const char *s)
{
	const char *c = s;
	char *stripped = (char*)calloc(sizeof(char),strlen(s)+1);
	size_t i = 0;
	while (*c != '\0')
	{
		if (*c == '%')
		{
			if (*(c + 1) == '\0')
			{
				break;
			}
			else
			{
				stripped[i] = *(c + 1);
				i++;
				c += 2;
			}
		}
		else
		{
			stripped[i] = *c;
			i++;
			c++;
		}
	}
	stripped[i] = '\0';
	return stripped;
}

char *
xrew_get_quoted(const char *s)
{
	const char *qstart = strchr((char*) s, '"') + 1;
	const char *qend = strrchr((char*) s, '"');
	char* qpart = HWFST::string_copy(qstart);
	*(qpart+ (size_t) (qend - qstart)) = '\0';
	return qpart;
}

HWFST::KeyPairSet*
_xrew_creation_pi()
{
	HWFST::KeyPairSet* pi = new HWFST::KeyPairSet;
	for (HWFST::KeySet::const_iterator k = _xrew_creation_sigma->begin();
			k != _xrew_creation_sigma->end();
			++k)
	{
		for (HWFST::KeySet::const_iterator k2 = _xrew_creation_sigma->begin();
				k2 != _xrew_creation_sigma->end();
				++k2)
		{
			pi->insert(HWFST::define_keypair(*k, *k2));
		}
	}
	return pi;
}

HWFST::KeyPairSet*
_xrew_negation_pi()
{
	HWFST::KeyPairSet* pi = new HWFST::KeyPairSet;
	for (HWFST::KeySet::const_iterator k = _xrew_negation_sigma->begin();
			k != _xrew_negation_sigma->end();
			++k)
	{
		for (HWFST::KeySet::const_iterator k2 = _xrew_negation_sigma->begin();
				k2 != _xrew_negation_sigma->end();
				++k2)
		{
			pi->insert(HWFST::define_keypair(*k, *k2));
		}
	}
	return pi;
}

HWFST::Key
_xrew_string_to_key(char *s)
{
	Key rv;
	if (HWFST::is_symbol(s))
	{
		rv = HWFST::get_key(HWFST::get_symbol(s), _xrew_key_table);
	}
	else
	{
		HWFST::Symbol sy = HWFST::define_symbol(s);
		rv = HWFST::get_unused_key(_xrew_key_table);
		HWFST::associate_key(rv, _xrew_key_table, sy);
	}
	return rv;
}

void
_xrew_new_key(HWFST::Key k)
{
	if (_xrew_keys->find(k) == _xrew_keys->end())
	{
		HWFST::insert_key(k, _xrew_keys);
		HWFST::insert_key(k, _xrew_creation_sigma);
		HWFST::insert_key(k, _xrew_negation_sigma);
	}
}

void
_xrew_new_transducer(HWFST::TransducerHandle t)
{
	HWFST::KeySet* keys = HWFST::define_key_set(t);
	for (HWFST::KeySet::const_iterator k = keys->begin();
		 k != keys->end(); ++k)
	{
		_xrew_new_key(*k);
	}
}

// XRE data structures
void
_xrew_new_data(const char* xrew)
{
	if (!_xrew_initialised)
	{
		_xrew_initialise();
	}
	_xrew_transducer_ = HWFST::create_epsilon_transducer();
	_xrew_data = strdup(xrew);
	_xrew_len = strlen(_xrew_data);
}

void _xrew_ensure_zero_epsilon()
{
	// FIXME
}

void _xrew_initialise(const char* xrew, HWFST::KeyPairSet* negation_sigma,
					 HWFST::KeyPairSet* creation_sigma, HWFST::KeyTable* keytable,
					 bool weighted)
{
	_xrew_transducer_ = (HWFST::TransducerHandle)malloc(sizeof(HWFST::TransducerHandle));
	_xrew_data = strdup(xrew);
	_xrew_len = strlen(_xrew_data);
	_xrew_negation_sigma = new HWFST::KeySet();
	for (HWFST::KeyPairSet::const_iterator kp= negation_sigma->begin();
			kp != negation_sigma->end();
			++kp)
	{
		_xrew_negation_sigma->insert(HWFST::get_input_key(*kp));
		_xrew_negation_sigma->insert(HWFST::get_output_key(*kp));
	}
	_xrew_creation_sigma = new HWFST::KeySet();
	for (HWFST::KeyPairSet::const_iterator kp= creation_sigma->begin();
			kp != creation_sigma->end();
			++kp)
	{
		_xrew_creation_sigma->insert(HWFST::get_input_key(*kp));
		_xrew_creation_sigma->insert(HWFST::get_output_key(*kp));
	}
	_xrew_keys = HWFST::create_empty_key_set();
	_xrew_key_table = keytable;
	_xrew_ensure_zero_epsilon();
	_xrew_weighted = false;
	_xrew_initialised = true;
}

void _xrew_initialise()
{
	_xrew_transducer_ = (HWFST::TransducerHandle)malloc(sizeof(HWFST::TransducerHandle));
	_xrew_data = (char*)calloc(sizeof(char), 1);
	_xrew_len = 0;
	_xrew_negation_sigma = HWFST::create_empty_key_set();
	_xrew_creation_sigma = HWFST::create_empty_key_set();
	_xrew_keys = HWFST::create_empty_key_set();
	//
	_xrew_key_table = HWFST::create_key_table();
	HWFST::Symbol zeroeps = HWFST::define_symbol("@0@");
	HWFST::associate_key(0, _xrew_key_table, zeroeps);
	_xrew_ensure_zero_epsilon();
	_xrew_weighted = false;
	_xrew_initialised = true;
}

HWFST::TransducerHandle
_xrew_make_key_pair(HWFST::Key k1, HWFST::Key k2)
{
	HWFST::TransducerHandle rv = (HWFST::TransducerHandle)malloc(sizeof(HWFST::TransducerHandle));
	if ( (k1 == ANY_KEY) && (k2 == ANY_KEY))
	{
		// any id pair
		HWFST::KeyPairSet* id_set = HWFST::create_empty_keypair_set();
		for (KeySet::const_iterator ck = _xrew_creation_sigma->begin();
			 ck != _xrew_creation_sigma->end(); ++ck)
		{
			id_set->insert(HWFST::define_keypair(*ck, *ck));
		}
		rv = HWFST::define_transducer(id_set);
	}
	else if ( k1 == ANY_KEY )
	{
		// ?:k2
		HWFST::KeyPairSet* any_set = HWFST::create_empty_keypair_set();
		for (KeySet::const_iterator ck = _xrew_creation_sigma->begin();
			 ck != _xrew_creation_sigma->end(); ++ck)
		{
				any_set->insert(HWFST::define_keypair(*ck, k2));
		}
		rv = HWFST::define_transducer(any_set);
	}
	else if ( k2 == ANY_KEY )
	{
		// k1:?
		HWFST::KeyPairSet* any_set = HWFST::create_empty_keypair_set();
		for (KeySet::const_iterator ck = _xrew_creation_sigma->begin();
			 ck != _xrew_creation_sigma->end(); ++ck)
		{
			any_set->insert(HWFST::define_keypair(k1, *ck));
		}
		rv = HWFST::define_transducer(any_set);
	}
	else
	{
		// regular key pair
		HWFST::KeyPair* kp = HWFST::define_keypair(k1, k2);
		rv = HWFST::define_transducer(kp);
	}
	return rv;
}

